package com.orbbec.orbbecsdkexamples.activity;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;

import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.os.Bundle;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;
import android.widget.Toast;

import com.orbbec.obsensor.AccelFrame;
import com.orbbec.obsensor.AccelStreamProfile;
import com.orbbec.obsensor.Device;
import com.orbbec.obsensor.DeviceChangedCallback;
import com.orbbec.obsensor.DeviceList;
import com.orbbec.obsensor.Frame;
import com.orbbec.obsensor.FrameCallback;
import com.orbbec.obsensor.FrameType;
import com.orbbec.obsensor.GyroFrame;
import com.orbbec.obsensor.GyroStreamProfile;
import com.orbbec.obsensor.OBContext;
import com.orbbec.obsensor.Sensor;
import com.orbbec.obsensor.SensorType;
import com.orbbec.obsensor.StreamProfileList;
import com.orbbec.obsensor.StreamType;
import com.orbbec.orbbecsdkexamples.R;

public class ImuActivity extends AppCompatActivity implements SurfaceHolder.Callback {
    private static final String TAG = "ImuActivity";
    private OBContext mOBContext;

    private Device mDevice;
    private Sensor mSensorAccel;
    private Sensor mSensorGyro;
    private AccelStreamProfile mAccelStreamProfile;
    private GyroStreamProfile mGyroStreamProfile;

    private SurfaceView mSurfaceViewImu;
    private Paint mPaintIMU = new Paint();
    private boolean mIsRefreshIMUDataRunning = false;
    private Thread mRefreshIMUDataThread;

    private AccelFrame mAccelFrame;
    private GyroFrame mGyroFrame;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_imu);
        mSurfaceViewImu = findViewById(R.id.sv_imuView);
        mSurfaceViewImu.getHolder().addCallback(this);
        // 1.初始化SDK, 并监听设备变化
        mOBContext = new OBContext(getApplicationContext(), new DeviceChangedCallback() {
            @Override
            public void onDeviceAttach(DeviceList deviceList) {
                try {
                    if (deviceList == null || deviceList.getDeviceCount() == 0) {
                        showToast("请接入设备");
                    } else {
                        // 2.创建Device
                        mDevice = deviceList.getDevice(0);

                        // 3.通过Device获取加速度Sensor
                        mSensorAccel = mDevice.getSensor(SensorType.ACCEL);

                        // 4.通过Device获取陀螺仪Sensor
                        mSensorGyro = mDevice.getSensor(SensorType.GYRO);

                        if (mSensorAccel == null || mSensorGyro == null) {
                            showToast("本设备不支持IMU");
                            return;
                        } else {
                            runOnUiThread(() -> {
                                mSurfaceViewImu.setVisibility(View.VISIBLE);
                            });
                        }

                        if (mSensorAccel != null && mSensorGyro != null) {
                            // 5.获取加速度计配置
                            StreamProfileList accelProfileList = mSensorAccel.getStreamProfileList();
                            if (null != accelProfileList) {
                                mAccelStreamProfile = accelProfileList.getStreamProfile(0).as(StreamType.ACCEL);
                                accelProfileList.close();
                            }

                            // 6.获取陀螺仪配置
                            StreamProfileList gyroProfileList = mSensorGyro.getStreamProfileList();
                            if (null != gyroProfileList) {
                                mGyroStreamProfile = gyroProfileList.getStreamProfile(0).as(StreamType.GYRO);
                                gyroProfileList.close();
                            }
                        }
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                } finally {
                    // 8.释放设备列表资源
                    deviceList.close();
                }
            }

            @Override
            public void onDeviceDetach(DeviceList deviceList) {
                try {
                    showToast("请接入设备");
                    deviceList.close();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        });
    }

    private void startAccelStream() {
        try {
            // 开启加速度计采样
            if (null != mAccelStreamProfile) {
                mSensorAccel.start(mAccelStreamProfile, new FrameCallback() {
                    @Override
                    public void onFrame(Frame frame) {
                        AccelFrame accelFrame = frame.as(FrameType.ACCEL);

                        if (null == mAccelFrame) {
                            mAccelFrame = accelFrame;
                            return;
                        }

                        frame.close();
                    }
                });
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void startGyroStream() {
        try {
            // 开启陀螺仪采样
            if (null != mGyroStreamProfile) {
                mSensorGyro.start(mGyroStreamProfile, new FrameCallback() {
                    @Override
                    public void onFrame(Frame frame) {
                        GyroFrame gyroFrame = frame.as(FrameType.GYRO);

                        if (null == mGyroFrame) {
                            mGyroFrame = gyroFrame;
                            return;
                        }

                        frame.close();
                    }
                });
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void startIMU() {
        // 7.1.初始化IMU数据刷新线程
        mIsRefreshIMUDataRunning = true;
        mRefreshIMUDataThread = new Thread(mRefreshIMUDataRunnable);
        mRefreshIMUDataThread.setName("RefreshIMUDataThread");

        // 7.2.开启IMU数据刷新线程
        mRefreshIMUDataThread.start();

        // 7.3.开始陀螺仪采样
        startGyroStream();

        // 7.4.开始加速度计采样
        startAccelStream();
    }

    private void stopIMU() {
        try {
            // 停止加速度计采样
            if (null != mSensorAccel) {
                mSensorAccel.stop();
            }

            // 停止陀螺仪采样
            if (null != mSensorGyro) {
                mSensorGyro.stop();
            }

            // 停止IMU数据刷新线程并释放
            mIsRefreshIMUDataRunning = false;
            if (null != mRefreshIMUDataThread) {
                try {
                    mRefreshIMUDataThread.join(300);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                mRefreshIMUDataThread = null;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void showToast(String msg) {
        runOnUiThread(() -> Toast.makeText(ImuActivity.this, msg, Toast.LENGTH_SHORT).show());
    }

    private Runnable mRefreshIMUDataRunnable = () -> {
        while (mIsRefreshIMUDataRunning) {
            drawImuInfo();
        }
    };

    private void drawImuInfo() {
        try {
            Canvas canvas = mSurfaceViewImu.getHolder().lockCanvas();
            if (null != canvas) {
                mPaintIMU.setTextSize(25.0f);
                Paint.FontMetrics fm = mPaintIMU.getFontMetrics();
                float offsetY = fm.descent - fm.ascent;
                float x = canvas.getWidth() / 8;
                float y = canvas.getHeight() / 8;
                canvas.drawColor(Color.WHITE);
                if (mAccelFrame != null) {
                    canvas.drawText("AccelTimestamp:" + mAccelFrame.getTimeStamp(), x, y, mPaintIMU);
                    canvas.drawText("AccelTemperature:" + mAccelFrame.getTemperature() + "dC", x, (y += offsetY), mPaintIMU);

                    canvas.drawText("Accel.x:" + mAccelFrame.getAccelData()[0] + "g", x, (y += 2 * offsetY), mPaintIMU);
                    canvas.drawText("Accel.y:" + mAccelFrame.getAccelData()[1] + "g", x, (y += offsetY), mPaintIMU);
                    canvas.drawText("Accel.z:" + mAccelFrame.getAccelData()[2] + "g", x, (y += offsetY), mPaintIMU);

                    // 释放AccelFrame资源
                    mAccelFrame.close();
                    mAccelFrame = null;
                }

                if (mGyroFrame != null) {
                    canvas.drawText("GyroTimestamp:" + mGyroFrame.getTimeStamp(), x, (y += 2 * offsetY), mPaintIMU);
                    canvas.drawText("GyroTemperature:" + mGyroFrame.getTemperature() + "dC", x, (y += offsetY), mPaintIMU);

                    canvas.drawText("Gyro.x:" + mGyroFrame.getGyroData()[0] + "dps", x, (y += 2 * offsetY), mPaintIMU);
                    canvas.drawText("Gyro.y:" + mGyroFrame.getGyroData()[1] + "dps", x, (y += offsetY), mPaintIMU);
                    canvas.drawText("Gyro.z:" + mGyroFrame.getGyroData()[2] + "dps", x, (y += offsetY), mPaintIMU);

                    // 释放GyroFrame资源
                    mGyroFrame.close();
                    mGyroFrame = null;
                }
                mSurfaceViewImu.getHolder().unlockCanvasAndPost(canvas);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    public void surfaceCreated(@NonNull SurfaceHolder holder) {
        // 开流
        startIMU();
    }

    @Override
    public void surfaceChanged(@NonNull SurfaceHolder holder, int format, int width, int height) {

    }

    @Override
    public void surfaceDestroyed(@NonNull SurfaceHolder holder) {

    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        try {

            stopIMU();

            // 释放加速度计配置
            if (null != mAccelStreamProfile) {
                mAccelStreamProfile.close();
                mAccelStreamProfile = null;
            }

            // 释放陀螺仪配置
            if (null != mGyroStreamProfile) {
                mGyroStreamProfile.close();
                mGyroStreamProfile = null;
            }

            // 释放Device
            if (null != mDevice) {
                mDevice.close();
                mDevice = null;
            }

            // 释放SDK
            if (null != mOBContext) {
                mOBContext.close();
                mOBContext = null;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}